import numpy as np
import scipy as sp


class TiseSolverFgh2D:
    def __init__(
        self,
        xa: float,
        xb: float,
        nx: int,
        ya: float,
        yb: float,
        ny: int,
        num_eig: int,
    ):
        self.xa: float = xa
        self.xb: float = xb
        self.nx: int = nx
        x = np.linspace(xa, xb, nx, endpoint=False, dtype=np.float64)
        self.x: np.ndarray = x

        self.ya: float = ya
        self.yb: float = yb
        self.ny: int = ny
        y = np.linspace(ya, yb, ny, endpoint=False, dtype=np.float64)
        self.y: np.ndarray = y

        X, Y = np.meshgrid(x, y, indexing="ij")  # X, Y has shapes (len(x), len(y))
        self.X = X
        self.Y = Y
        self.Vxy: np.ndarray = self.V(X, Y)

        self.num_eig: int = num_eig
        self.eigenvalues: np.ndarray = np.zeros(self.num_eig, dtype=np.complex128)
        self.eigenvectors: np.ndarray = np.zeros((self.num_eig, self.nx, self.ny), dtype=np.complex128)

    # TISE potential
    def V(self, x: np.ndarray, y: np.ndarray):
        V = 0.5 * x * x + 0.5 * y * y
        return V

    def abscissa(self):
        return self.x, self.y, self.X, self.Y

    def solve(self):
        # Hamiltonian matrix (N x N)
        matH_sparse = TiseSolverFgh2D.matH(self.x, self.y, self.Vxy)

        print("Diagonalize")
        # Diagonalize
        is_hermite: bool = False
        is_full_diagonal: bool = self.num_eig == self.nx
        if is_full_diagonal:
            matH = matH_sparse.toarray()  # sparse matrix to density matrix
            if is_hermite:
                eigenvalues, eigenvectors = sp.linalg.eig(matH, left=False, right=True)
            else:
                eigenvalues, eigenvectors = sp.linalg.eigh(matH, left=False, right=True)
        else:
            if is_hermite:
                eigenvalues, eigenvectors = sp.sparse.linalg.eigsh(matH_sparse, k=self.num_eig, which="SA")
            else:
                eigenvalues, eigenvectors = sp.sparse.linalg.eigs(matH_sparse, k=self.num_eig, which="SR")
        print("Diagonalize end")

        # Sort eigenvalues and eigenvectors in assending order by real part of the eigenvalues.
        idx = np.argsort(eigenvalues.real)
        eigenvalues = eigenvalues[idx]
        eigenvectors = eigenvectors[:, idx]

        # Normalize to \int |\psi(x)|^2 dx = 1, (Just in case)
        S = np.sum(np.abs(eigenvectors) ** 2, axis=0)
        eigenvectors /= np.sqrt(S)

        # Eigenvector to eigenfunction
        # eigenfunctions = eigenvectors.astype(np.complex128) * np.sqrt(self.nx / (self.xb - self.xa)) * np.sqrt(self.ny / (self.yb - self.ya))
        # eigenfunctions_reshape = np.stack([eigenfunctions[:, i].reshape(self.ny, self.nx).T for i in range(self.num_eig)], axis=-1)

        self.eigenvalues = eigenvalues.copy()
        self.eigenvectors = (np.stack([eigenvectors[:, i].reshape(self.ny, self.nx).T for i in range(self.num_eig)], axis=0)).copy()

        return eigenvalues  # .real, eigenfunctions_reshape.real

    @staticmethod
    def T2N_legacy(n, N):
        if N % 2 == 0:
            if n == 0:
                T = (1.0 / 12.0) * (N * N + 2.0)
            else:
                T = 0.5 * (-1.0) ** n / ((np.sin(np.pi * n / N)) ** 2)
        else:
            if n == 0:
                T = (1.0 / 12.0) * (N**2 - 1.0)
            else:
                T = 0.5 * (-1.0) ** n * np.cos(np.pi * n / N) / ((np.sin(np.pi * n / N)) ** 2)

        return T

    @staticmethod
    def T2N(n, N):
        """
        ベクトル n (整数値の配列) に対して T2N(n, N) を要素ごとに計算する関数
        """
        n = np.asarray(n)
        result = np.empty_like(n, dtype=np.float64)
        mask = n == 0
        if N % 2 == 0:
            T0 = (1.0 / 12.0) * (N * N + 2.0)
            result[mask] = T0
            nonzero = ~mask
            result[nonzero] = 0.5 * ((-1) ** np.abs(n[nonzero])) / (np.sin(np.pi * n[nonzero] / N) ** 2)
        else:
            T0 = (1.0 / 12.0) * (N**2 - 1)
            result[mask] = T0
            nonzero = ~mask
            result[nonzero] = 0.5 * ((-1) ** np.abs(n[nonzero])) * np.cos(np.pi * n[nonzero] / N) / (np.sin(np.pi * n[nonzero] / N) ** 2)
        return result

    @staticmethod
    def matH(
        x: np.ndarray,
        y: np.ndarray,
        V: np.ndarray,
        is_code_optimization: bool = True,
    ):
        """
        x, y: 1次元配列。グリッド点
        V: ポテンシャル配列 (shape=(Nx, Ny))。V[n, m] が (x_n, y_m) におけるポテンシャル
        """
        Nx = x.shape[0]  # V.shape[0] と同じ
        Ny = y.shape[0]  # V.shape[1] と同じ
        N = Nx * Ny

        dx = x[1] - x[0]
        dy = y[1] - y[0]
        Kx = 2.0 * np.pi / dx
        Ky = 2.0 * np.pi / dy

        if is_code_optimization:
            # x方向の運動エネルギー項: ブロック対角行列 np.kron(np.eye(Ny), T_x)
            n_idx = np.arange(Nx)
            diff_x = n_idx[:, None] - n_idx[None, :]  # shape: (Nx, Nx)
            T_x = TiseSolverFgh2D.T2N(diff_x, Nx)
            T_x = 0.5 * (Kx / Nx) ** 2 * T_x

            # y方向の運動エネルギー項: np.kron(T_y, np.eye(Nx))
            m_idx = np.arange(Ny)
            diff_y = m_idx[:, None] - m_idx[None, :]  # shape: (Ny, Ny)
            T_y = TiseSolverFgh2D.T2N(diff_y, Ny)
            T_y = 0.5 * (Ky / Ny) ** 2 * T_y

            # スパース行列に変換
            I_Ny = sp.sparse.eye(Ny, format="csr")
            I_Nx = sp.sparse.eye(Nx, format="csr")
            T_x_sparse = sp.sparse.csr_matrix(T_x)
            T_y_sparse = sp.sparse.csr_matrix(T_y)

            # sparseなKronecker積を用いて全体の運動エネルギー行列を構成
            ham = sp.sparse.kron(I_Ny, T_x_sparse, format="csr") + sp.sparse.kron(T_y_sparse, I_Nx, format="csr")

            # ポテンシャル項（対角成分のみ）の追加
            ham = ham + sp.sparse.diags(V.T.ravel(), format="csr")

        else:
            # Initialize the Hamiltonian matrix
            ham = np.zeros((N, N), dtype=np.float64)

            rows = []
            cols = []
            data = []

            for m1 in range(Ny):
                for n1 in range(Nx):
                    i = m1 * Nx + n1
                    for m2 in range(Ny):
                        for n2 in range(Nx):
                            j = m2 * Nx + n2

                            t1 = 0.0
                            if m1 == m2:
                                t1 = 0.5 * TiseSolverFgh2D.T2N_legacy(n1 - n2, Nx) * (Kx / Nx) ** 2

                            t2 = 0.0
                            if n1 == n2:
                                t2 = 0.5 * TiseSolverFgh2D.T2N_legacy(m1 - m2, Ny) * (Ky / Ny) ** 2

                            t3 = 0.0
                            if n1 == n2 and m1 == m2:
                                t3 = V[n1, m1]

                            value = t1 + t2 + t3
                            if value != 0.0:
                                rows.append(i)
                                cols.append(j)
                                data.append(value)
            ham = sp.sparse.csr_matrix((data, (rows, cols)), shape=(N, N))
        return ham

    def eigenfunctions_scaling(self, x_new: np.ndarray, y_new: np.ndarray):
        if self.X.shape == x_new.shape and self.Y.shape == y_new.shape:
            rtol = atol = 10.0 * np.finfo(np.float64).eps
            if np.allclose(self.X, x_new, rtol=rtol, atol=atol) and np.allclose(self.Y, y_new, rtol=rtol, atol=atol):
                # abscissa is same as fgh-grid
                eigenfunctions = self.eigenvectors * np.sqrt(self.nx / (self.xb - self.xa)) * np.sqrt(self.ny / (self.yb - self.ya))
                return eigenfunctions

        # Interpolation using FGH basis
        Lx = self.xb - self.xa
        Ly = self.yb - self.ya

        # --- x-direction basis functions ---
        if self.nx % 2 == 0:
            l_idx_x = np.arange(-self.nx // 2, self.nx // 2)  # shape: (nx,)
        else:
            l_idx_x = np.arange(-(self.nx - 1) // 2, (self.nx - 1) // 2 + 1)
        n_vals = np.arange(self.nx)  # basis index n, shape: (nx,)
        n_over = n_vals / self.nx  # shape: (nx,)
        # Normalized x-coordinates (shape: (len(x_new),))
        A = (x_new - self.xa) / Lx
        # For each n, compute A - n/nx, shape: (nx, len(x_new))
        temp_x = A[None, :] - n_over[:, None]
        # Compute exponent for each n, x, and l, resulting shape: (nx, len(x_new), nx)
        # l_idx_x[None, None, :] has shape: (1, 1, nx)
        exponent_x = 2j * np.pi * temp_x[..., None] * l_idx_x[None, None, :]
        # Sum over the l-direction for each n and x, resulting shape: (nx, len(x_new))
        varphi_x = np.sum(np.exp(exponent_x), axis=2) / np.sqrt(self.nx * Lx)

        # --- y-direction basis functions ---
        if self.ny % 2 == 0:
            l_idx_y = np.arange(-self.ny // 2, self.ny // 2)
        else:
            l_idx_y = np.arange(-(self.ny - 1) // 2, (self.ny - 1) // 2 + 1)
        m_vals = np.arange(self.ny)  # basis index m, shape: (ny,)
        m_over = m_vals / self.ny  # shape: (ny,)
        # Normalized y-coordinates (shape: (len(y_new),))
        B = (y_new - self.ya) / Ly
        # For each m, compute B - m/ny, shape: (ny, len(y_new))
        temp_y = B[None, :] - m_over[:, None]
        # Compute exponent for each m, y, and l, resulting shape: (ny, len(y_new), ny)
        exponent_y = 2j * np.pi * temp_y[..., None] * l_idx_y[None, None, :]
        # Sum over the l-direction for each m and y, resulting shape: (ny, len(y_new))
        varphi_y = np.sum(np.exp(exponent_y), axis=2) / np.sqrt(self.ny * Ly)

        # Wavefunction: wf(k, x, y) = sum_{n,m} ev[k, n, m] * varphi_x[n, x] * varphi_y[m, y]
        eigenfunctions = np.einsum("knm,nx,my->kxy", self.eigenvectors, varphi_x, varphi_y)
        return eigenfunctions
